unit PortServers;

interface

uses
  Windows, Classes, DServerSocks;

const
  //`@
  InDataMask       = Word($0001);
  InEnableMask     = Word($0100);
  OutDataMask      = Word($0001);
  //OutEnableMask    = Word($0004);
  //OutNotEnableMask = $FFFF xor OutEnableMask;
  DiscardDepth = 400;

type
  TPortServer = class;

  PPacketData = ^TPacketData;
  TPacketData = record
    Prev, Next: PPacketData;
    Timestamp: TDateTime;
    LenUsed: LongWord;
    Buf: array of Byte;
  end;

  TPacketFIFO = class
  public
    Inlet, Outlet: PPacketData;
    Depth: Integer;
    hListMutex, hListSem: THandle;
    bShutdown: Boolean;
    constructor Create; virtual;
    destructor Destroy; override;
    procedure Insert(NewPacket: PPacketData);
    function Extract(Timeout: LongWord = 0): PPacketData;
    function ExtractAll: PPacketData;
    procedure Clear;
    function PeekDepth: Integer;
  end;



  TClientData = record
    pSC: TDSSServerClient;
    InBuf: String;
  end;
  PClientData = ^TClientData;

  TPortServer = class(TThread)
  protected
    procedure Execute; override;
    procedure ExecuteDIOWorker; virtual; abstract;
    procedure ReplyNotWorking(pSC: TDSSServerClient); virtual;
    procedure ReplyWorking(pSC: TDSSServerClient); virtual;

  public
    DeviceIndex: LongWord;
    bRead: Boolean;

    Status: LongWord;
    bWorking, bRunning, bRejectData: Boolean;

    ServerSock: TDServerSock;
    ClientData: array of PClientData;
    bGotData: Boolean;

    PacketFIFO: TPacketFIFO;
    hClientDataMutex, hRejectedSem: THandle;
    MaxDIOBuf: Integer;
    PacketBits: Integer;
    //LastOverflow: TDateTime;
    LastHz: Integer;
    LastTry: String;
    iPacketState: Integer;
    constructor Create; virtual;
    destructor Destroy; override;
    procedure Init(EID: LongWord; bIsRead: Boolean); virtual;
    procedure SetFreq(Hz: Integer); virtual;
    function TryStartBoard: LongWord; virtual;
    procedure DoClear; virtual;
    procedure DoCleanup; virtual;
    procedure CheckData; virtual;
    procedure ServerSockClientConnect(Sender: TObject; ServerClient: TDSSServerClient); virtual;
    procedure ServerSockClientGotData(Sender: TObject; ServerClient: TDSSServerClient); virtual;
    procedure ServerSockClientDisconnect(Sender: TObject; ServerClient: TDSSServerClient); virtual;
  end;

  TReadPortServer = class(TPortServer)
  protected
    procedure ExecuteDIOWorker; override;
  end;

  TWritePortServer = class(TPortServer)
  protected
    procedure ExecuteDIOWorker; override;
  end;

var
  bNeedsCleanup: Boolean;





implementation

uses
  SysUtils, AIOUSB, DIFFChunks, ServUnit;

function IntAsStr(I: LongWord): String;
begin
  SetString(Result, PChar(@I), 4);
end;

function StrAsInt(S: String): LongWord;
begin
  case Length(S) of
    0: Result := 0;
    1: Result := PByte(@S[1])^;
    2: Result := PWord(@S[1])^;
    else Result := PLongWord(@S[1])^; //3 works because non-empty strings are null-terminated
  end;
end;



{ TPacketFIFO }

procedure TPacketFIFO.Clear;
var
  ThisPacket, LastPacket: PPacketData;

begin
  WaitForSingleObject(hListMutex, INFINITE);
    ThisPacket := Outlet;
    while ThisPacket <> nil do begin
      WaitForSingleObject(hListSem, 0);

      LastPacket := ThisPacket;
      ThisPacket := ThisPacket.Next;
      Dispose(LastPacket);
    end;
    Inlet := nil;
    Outlet := nil;
    Depth := 0;
  ReleaseMutex(hListMutex);
end;

constructor TPacketFIFO.Create;
begin
  Inlet := nil;
  Outlet := nil;
  hListMutex := CreateMutex(nil, False, nil);
  hListSem := CreateSemaphore(nil, 0, $FFFF, nil);
end;

destructor TPacketFIFO.Destroy;
begin
  Clear;

  CloseHandle(hListMutex);
  CloseHandle(hListSem);

  inherited;
end;

function TPacketFIFO.Extract(Timeout: LongWord = 0): PPacketData;
begin
  if bShutdown then begin
    Result := nil;
    Exit;
  end;

  WaitForSingleObject(hListSem, Timeout);

  WaitForSingleObject(hListMutex, INFINITE);
    Result := Outlet;
    if Result <> nil then begin
      Outlet := Result.Next;
      if Outlet <> nil then
        Outlet.Prev := nil
      else
        Inlet := nil
      ;
      Dec(Depth);
    end;
  ReleaseMutex(hListMutex);
end;

function TPacketFIFO.ExtractAll: PPacketData;
var
  I: Integer;

begin
  WaitForSingleObject(hListMutex, INFINITE);
    Result := Outlet;
    Outlet := nil;
    Inlet := nil;

    I := Depth;
    Depth := 0;

    for I := 1 to I do
      WaitForSingleObject(hListSem, 0)
    ;
  ReleaseMutex(hListMutex);
end;

procedure TPacketFIFO.Insert(NewPacket: PPacketData);
var
  bExtract: Boolean;
  ThatPacket: PPacketData;

begin
  WaitForSingleObject(hListMutex, INFINITE);
    NewPacket.Next := nil;
    NewPacket.Prev := Inlet;
    if Inlet = nil then Outlet := NewPacket else Inlet.Next := NewPacket;
    Inlet := NewPacket;

    Inc(Depth);
    bExtract := Depth > DiscardDepth;
  ReleaseMutex(hListMutex);

  ReleaseSemaphore(hListSem, 1, nil);

  if bExtract then begin
    ThatPacket := Extract;
    if ThatPacket <> nil then Dispose(ThatPacket);
  end;
end;

function TPacketFIFO.PeekDepth: Integer;
begin
  WaitForSingleObject(hListMutex, INFINITE);
    Result := Depth;
  ReleaseMutex(hListMutex);
end;

{ TPortServer }

procedure TPortServer.CheckData;
var
  I, L: Integer;
  GatheredClientData: array of record
    pSC: TDSSServerClient;
    ChunkStr: String;
  end;
  pInChunk, pOutChunk, pErrChunk: TDIFFChunk;
  Buf, Timestamps: String;
  ThisPacket, PrevPacket: PPacketData;

begin
  if not bGotData then Exit;

  WaitForSingleObject(hClientDataMutex, INFINITE);
    bGotData := False;

    L := Length(ClientData);
    SetLength(GatheredClientData, L);
    for I := 0 to L-1 do begin
      GatheredClientData[I].pSC := ClientData[I].pSC;

      try
        GatheredClientData[I].ChunkStr := ExtractChunkStr(ClientData[I].InBuf);
      except
        on E: Exception do begin
          GatheredClientData[I].ChunkStr := '';
          //RequestResendForException(ClientData[I].pSC, E);
        end;
      end;
    end;
  ReleaseMutex(hClientDataMutex);

  for I := 0 to L-1 do if Length(GatheredClientData[I].ChunkStr) <> 0 then begin
    bGotData := True;
    pInChunk := nil;
    try
      pInChunk := TDIFFChunk.CreateFromString(GatheredClientData[I].ChunkStr);

      if pInChunk.TagStr = 'RREQL' then begin
        if not bWorking then begin
          ReplyNotWorking(GatheredClientData[I].pSC);
        end
        else if bRead then begin
          pOutChunk := TDIFFChunk.Create;
          pOutChunk.TagStr := 'RDATA';

          ThisPacket := PacketFIFO.ExtractAll;
          if ThisPacket = nil then
            pOutChunk.SpecData := ''
          else begin
            while ThisPacket.Next <> nil do begin
              PrevPacket := ThisPacket;
              ThisPacket := ThisPacket.Next;
              Dispose(PrevPacket);
            end;
            SetString(pOutChunk.SpecData, PChar(@ThisPacket.Buf[0]), ThisPacket.LenUsed);
            SetString(Timestamps, PChar(@ThisPacket.Timestamp), SizeOf(ThisPacket.Timestamp));
            pOutChunk.AddHeader('Time', Timestamps);
            Dispose(ThisPacket);
          end;
          Buf := pOutChunk.SaveToString;
          FreeAndNil(pOutChunk);
          LastTry := Buf;
          GatheredClientData[I].pSC.PostSendStr(Buf);
        end
        else begin
          pErrChunk := TDIFFChunk.Create;
          pErrChunk.TagStr := 'IOERR';
          pErrChunk.AddHeader('Type', 'No Read');
          GatheredClientData[I].pSC.PostSendStr(pErrChunk.SaveToString);
          FreeAndNil(pErrChunk);
        end;
      end
      else if pInChunk.TagStr = 'RREQA' then begin
        if not bWorking then begin
          ReplyNotWorking(GatheredClientData[I].pSC);
        end
        else if bRead then begin
          pOutChunk := TDIFFChunk.Create;
          pOutChunk.TagStr := 'RDATA';

          ThisPacket := PacketFIFO.ExtractAll;
          Timestamps := '';
          pOutChunk.SpecData := '';
          if ThisPacket <> nil then begin
            L := (PacketBits + 7) div 8;
            repeat
              SetString(Buf, PChar(@ThisPacket.Timestamp), SizeOf(ThisPacket.Timestamp));
              Timestamps := Timestamps + Buf;

              SetString(Buf, PChar(@ThisPacket.Buf[0]), ThisPacket.LenUsed);
              if Length(Buf) > L then SetLength(Buf, L) else if Length(Buf) < L then Buf := Buf + StringOfChar(#0, L - Length(Buf));
              pOutChunk.SpecData := pOutChunk.SpecData + Buf;

              PrevPacket := ThisPacket;
              ThisPacket := ThisPacket.Next;
              Dispose(PrevPacket);
            until ThisPacket = nil;
            pOutChunk.AddHeader('Time', Timestamps);
            pOutChunk.AddHeader('Size', IntAsStr(L));
          end;
          Buf := pOutChunk.SaveToString;
          FreeAndNil(pOutChunk);
          LastTry := Buf;
          GatheredClientData[I].pSC.PostSendStr(Buf);
        end
        else begin
          pErrChunk := TDIFFChunk.Create;
          pErrChunk.TagStr := 'IOERR';
          pErrChunk.AddHeader('Type', 'No Read');
          GatheredClientData[I].pSC.PostSendStr(pErrChunk.SaveToString);
          FreeAndNil(pErrChunk);
        end;
      end
      else if pInChunk.TagStr = 'RREQ?' then begin
        if not bWorking then begin
          ReplyNotWorking(GatheredClientData[I].pSC);
        end
        else if bRead then begin
          pOutChunk := TDIFFChunk.Create;
          pOutChunk.TagStr := 'RDATA';

          ThisPacket := PacketFIFO.Extract(0);
          if ThisPacket = nil then begin
            pOutChunk.SpecData := '';
          end
          else begin
            SetString(pOutChunk.SpecData, PChar(@ThisPacket.Buf[0]), ThisPacket.LenUsed);
            Dispose(ThisPacket);
          end;
          Buf := pOutChunk.SaveToString;
          FreeAndNil(pOutChunk);
          LastTry := Buf;
          GatheredClientData[I].pSC.PostSendStr(Buf);
        end
        else begin
          pErrChunk := TDIFFChunk.Create;
          pErrChunk.TagStr := 'IOERR';
          pErrChunk.AddHeader('Type', 'No Read');
          GatheredClientData[I].pSC.PostSendStr(pErrChunk.SaveToString);
          FreeAndNil(pErrChunk);
        end;
      end
      else if pInChunk.TagStr = 'RETRY' then begin
        GatheredClientData[I].pSC.PostSendStr(LastTry);
      end
      else if pInChunk.TagStr = 'WDATA' then begin
        if not bWorking then begin
          ReplyNotWorking(GatheredClientData[I].pSC);
        end
        else if bRead then begin
          pErrChunk := TDIFFChunk.Create;
          pErrChunk.TagStr := 'IOERR';
          pErrChunk.AddHeader('Type', 'No Write');
          GatheredClientData[I].pSC.PostSendStr(pErrChunk.SaveToString);
          FreeAndNil(pErrChunk);
        end
        else begin
          Buf := pInChunk.SpecData;
          L := Length(Buf);
          if L > 0 then begin
            New(ThisPacket);
            ThisPacket.LenUsed := L;
            SetLength(ThisPacket.Buf, ThisPacket.LenUsed);
            Move(pInChunk.SpecData[1], ThisPacket.Buf[0], L);
            PacketFIFO.Insert(ThisPacket);
          end;
          Buf := '';

          pOutChunk := TDIFFChunk.Create;
          pOutChunk.TagStr := 'WRESP';
          pOutChunk.AddHeader('Bytes', IntAsStr(L));
          GatheredClientData[I].pSC.PostSendStr(pOutChunk.SaveToString);
          FreeAndNil(pOutChunk);
        end;
      end
      else if pInChunk.TagStr = 'CLEAR' then begin
        if not bWorking then begin
          //ReplyNotWorking(GatheredClientData[I].pSC);
        end
        else begin
          DoClear;
        end;
      end
      else if pInChunk.TagStr = 'GOCLR' then begin
        if not bWorking then begin
          //ReplyNotWorking(GatheredClientData[I].pSC);
        end
        else begin
          DoClear;
          if not bRunning then Resume;
        end;
      end
      else if pInChunk.TagStr = 'GOGO.' then begin
        if not bWorking then begin
          //ReplyNotWorking(GatheredClientData[I].pSC);
        end
        else begin
          if not bRunning then Resume;
        end;
      end
      else if pInChunk.TagStr = 'WFREQ' then begin
        if not bWorking then begin
          //ReplyNotWorking(GatheredClientData[I].pSC);
        end
        else begin
          SetFreq(StrAsInt(pInChunk.HeaderByName['Hz']));
        end;
      end
      else if pInChunk.TagStr = 'WBITS' then begin
        if not bWorking then begin
          //ReplyNotWorking(GatheredClientData[I].pSC);
        end
        else begin
          PacketBits := StrAsInt(pInChunk.HeaderByName['Bits']);
        end;
      end
      else if pInChunk.TagStr = 'WBKSZ' then begin

        L := StrAsInt(pInChunk.HeaderByName['Bytes']);
        if L < $200 then
          L := $200
        else if (L and $1FF) <> 0 then
          L := (L + $200) and not $1FF
        ;
        MaxDIOBuf := L;

      end
      else if pInChunk.TagStr = 'WDIO1' then begin

        for L := 0 to High(pInChunk.Header) do
          with pInChunk.Header[L] do
            if (Length(Nom) = 1) and (Nom[1] < #32) and (Length(Dat) = 1) then
              DIO_Write1(DeviceIndex, Ord(Nom[1]), Dat[1] <> #0)
        ;

      end
      else if pInChunk.TagStr = 'OKOK?' then begin
        if bWorking then
          ReplyWorking(GatheredClientData[I].pSC)
        else
          ReplyNotWorking(GatheredClientData[I].pSC)
        ;
      end;

    except
      //on E: Exception do RequestResendForException(GatheredClientData[I].pSC, E);
    end;
    FreeAndNil(pInChunk);
  end;
end;

constructor TPortServer.Create;
begin
  inherited Create(True);
  MaxDIOBuf := 512;
  PacketBits := 22*8;

  hClientDataMutex := CreateMutex(nil, False, nil);
  hRejectedSem := CreateSemaphore(nil, 0, 42, nil);

  PacketFIFO := TPacketFIFO.Create;

  ServerSock := TDServerSock.Create;
  ServerSock.OnClientConnect := ServerSockClientConnect;
  ServerSock.OnClientGotData := ServerSockClientGotData;
  ServerSock.OnClientDisconnect := ServerSockClientDisconnect;
end;

destructor TPortServer.Destroy;
begin
  ServerSock.Free;

  PacketFIFO.bShutdown := True;
  ReleaseSemaphore(PacketFIFO.hListSem, 10, nil);
  
  if bRunning then begin
    Terminate;
    WaitFor;
  end;

  PacketFIFO.Free;

  CloseHandle(hClientDataMutex);
  CloseHandle(hRejectedSem);

  inherited;
end;

procedure TPortServer.DoCleanup;
begin
  ServerSock.CleanupThreads;
  if ServerSock.CleanupListen <> -1 then
    ServerSock.GoListen
  ;
end;

procedure TPortServer.DoClear;
var
  EndTicks: LongWord;
  dTicks: Integer;
  Status: LongWord;

begin
  EndTicks := GetTickCount + 1000;
  repeat
    Status := WaitForSingleObject(hRejectedSem, 0);
    dTicks := GetTickCount - EndTicks;
  until (Status = WAIT_TIMEOUT) or (dTicks > 0);

  bRejectData := True;

  AIOUSB_ClearFIFO(DeviceIndex, 5);

  if bRunning then begin
    WaitForSingleObject(hRejectedSem, 1000);
  end;

  PacketFIFO.Clear;

  bRejectData := False;
end;

procedure TPortServer.Execute;
begin
  bRunning := True;
  
  ExecuteDIOWorker;

  if bWorking then
    Status := DIO_StreamClose(DeviceIndex)
  else
    DIO_StreamClose(DeviceIndex)
  ;
  bWorking := False;
  bRunning := False;
end;

procedure TPortServer.Init(EID: LongWord; bIsRead: Boolean);
begin
  bRead := bIsRead;

  DeviceIndex := GetDeviceByEEPROMData(0, SizeOf(EID), PByte(@EID));
  if DeviceIndex = diNone then begin
    Status := ERROR_FILE_NOT_FOUND;
    bWorking := False;
  end
  else begin
    Status := TryStartBoard;
    bWorking := Status = ERROR_SUCCESS;
  end;
  
  ServerSock.ListenPort := Word(EID);
  ServerSock.GoListen;
end;

procedure TPortServer.ReplyNotWorking(pSC: TDSSServerClient);
var
  pChunk: TDIFFChunk;

begin
  pChunk := TDIFFChunk.Create;
  pChunk.TagStr := 'NOWRK';
  pChunk.AddHeader('Status', IntAsStr(Status));
  pSC.PostSendStr(pChunk.SaveToString);
  FreeAndNil(pChunk);
end;

procedure TPortServer.ReplyWorking(pSC: TDSSServerClient);
var
  pChunk: TDIFFChunk;

begin
  pChunk := TDIFFChunk.Create;
  pChunk.TagStr := 'OKOK.';
  pChunk.AddHeader('Status', IntAsStr(Status));
  pSC.PostSendStr(pChunk.SaveToString);
  FreeAndNil(pChunk);
end;

procedure TPortServer.ServerSockClientConnect(Sender: TObject; ServerClient: TDSSServerClient);
var
  I: Integer;

begin
  WaitForSingleObject(hClientDataMutex, INFINITE);
  try
    for I := 0 to High(ClientData) do //Just in case.
      if ClientData[I].pSC = ServerClient then
        Exit
    ;

    I := Length(ClientData);
    SetLength(ClientData, I + 1);
    New(ClientData[I]);
    ClientData[I].pSC := ServerClient;
    ClientData[I].InBuf := '';

  finally
    ReleaseMutex(hClientDataMutex);
  end;
end;

procedure TPortServer.ServerSockClientDisconnect(Sender: TObject; ServerClient: TDSSServerClient);
var
  H, I: Integer;

begin
  bNeedsCleanup := True;

  WaitForSingleObject(hClientDataMutex, INFINITE);
  try
    H := High(ClientData);
    for I := 0 to H do
      if ClientData[I].pSC = ServerClient then begin
        Dispose(ClientData[I]);
        if I < H then ClientData[I] := ClientData[H];
        SetLength(ClientData, H);
        Break;
      end
    ;
  finally
    ReleaseMutex(hClientDataMutex);
  end;

  ServForm.DoPump;
end;

procedure TPortServer.ServerSockClientGotData(Sender: TObject; ServerClient: TDSSServerClient);
var
  NewBuf: String;
  I: Integer;
  bFound: Boolean;

begin
  NewBuf := ServerClient.ReceiveStr;

  WaitForSingleObject(hClientDataMutex, INFINITE);
  try
    bFound := False;
    for I := 0 to High(ClientData) do
      with ClientData[I]^ do
        if pSC = ServerClient then begin
          InBuf := InBuf + NewBuf;
          bFound := True; //Artificial lack of gotos considered harmful.
          Break;
        end
    ;
    if not bFound then begin //Just in case.
      I := Length(ClientData);
      SetLength(ClientData, I + 1);
      New(ClientData[I]);
      ClientData[I].pSC := ServerClient;
      ClientData[I].InBuf := NewBuf;
    end;
    bGotData := True;
  finally
    ReleaseMutex(hClientDataMutex);

    ServForm.DoPump;
  end;
end;

procedure TPortServer.SetFreq(Hz: Integer);
var
  ReadHz, WriteHz: Double;

begin
  if bRead then begin
    ReadHz := Hz;
    WriteHz := 0;
  end
  else begin
    ReadHz := 0;
    WriteHz := Hz;
  end;
  Status := DIO_StreamSetClocks(DeviceIndex, ReadHz, WriteHz);
  bWorking := Status = ERROR_SUCCESS;
  LastHz := Hz;
end;

function TPortServer.TryStartBoard: LongWord;
var
  ReadHz, WriteHz: Double;
  OutMask, TristateMask: Byte;
  DIOData: LongWord;

begin
  OutMask := $7;
  DIOData := $FFFFFFFF;
  TristateMask := 3;
  Result := DIO_ConfigureEx(DeviceIndex, @OutMask, @DIOData, @TristateMask);
  if Result <> ERROR_SUCCESS then Exit;

  ReadHz := 0;
  WriteHz := 0;
  Result := DIO_StreamSetClocks(DeviceIndex, ReadHz, WriteHz);
  if Result <> ERROR_SUCCESS then Exit;
  LastHz := 0;

  Result := DIO_StreamOpen(DeviceIndex, bRead);
  if Result <> ERROR_SUCCESS then Exit;

  TristateMask := 0;
  Result := DIO_ConfigureEx(DeviceIndex, @OutMask, @DIOData, @TristateMask);
end;

{ TReadPortServer }

procedure TReadPortServer.ExecuteDIOWorker;
var
  I, L, L2, BytesTransferred: LongInt;
  DIOBuf: array of Word;
  pDIOBuf: PPacketData;
  PacketL: LongWord;
  PacketState: (psBeforetime, psPacket, psDeadAir);
  UnbufferedByte: Byte;
  UnbufferedBits: Integer;

  procedure GetBuf;
  begin
    New(pDIOBuf);
    pDIOBuf.LenUsed := 0;
    SetLength(pDIOBuf.Buf, PacketL);
  end;

  procedure BufferByte;
  begin
    if pDIOBuf.LenUsed >= PacketL then begin
      PacketL := PacketL * 2;
      SetLength(pDIOBuf.Buf, PacketL);
    end;
    pDIOBuf.Buf[pDIOBuf.LenUsed] := UnbufferedByte;
    Inc(pDIOBuf.LenUsed);
    UnbufferedByte := 0;
    UnbufferedBits := 0;
  end;

begin
  pDIOBuf := nil;
  PacketState := psBeforetime;
  PacketL := 512;
  UnbufferedByte := 0;
  UnbufferedBits := 0;

  while not Terminated do begin
    L := MaxDIOBuf;
    L2 := L div 2;

    if Length(DIOBuf) <> L2 then begin
      SetLength(DIOBuf, 0);
      SetLength(DIOBuf, L2);
    end;

    BytesTransferred := 0;
    if bRejectData then begin
      Status := ERROR_SUCCESS;
      ReleaseSemaphore(hRejectedSem, 1, nil);
    end
    else begin
      Status := DIO_StreamFrame(DeviceIndex, L2, @DIOBuf[0], LongWord(BytesTransferred));
    end;
    if Status = ERROR_HANDLE_EOF then begin
      Status := ERROR_SUCCESS; //A sub-block partial. We're streaming, so we don't really care.
    end
    else if Status <> ERROR_SUCCESS then begin
      bWorking := False;
      Exit; //Fail, kick out. `@ Maybe should re-find board and reinit, depending on error code?
    end;
    if BytesTransferred < 0 then BytesTransferred := 0; //Just in case.
    if L > BytesTransferred then L := BytesTransferred;

    //for I := 0 to (L div 2) - 1 do DIOBuf[I] := DIOBuf[I] and (InEnableMask or InDataMask);

    for I := 0 to (L div 2) - 1 do begin
      if (DIOBuf[I] and InEnableMask) = 0 then begin //If enabled(active low).

        if PacketState = psBeforetime then begin
          //A partial. Give it up, we can't get the whole thing.
        end
        else begin
          if PacketState = psDeadAir then begin
            GetBuf;
            PacketState := psPacket;
          end;
          //Accumulate the data.
          if (DIOBuf[I] and InDataMask) <> 0 then UnbufferedByte := UnbufferedByte or (1 shl UnbufferedBits);
          Inc(UnbufferedBits);
          if UnbufferedBits >= 8 then BufferByte;
        end; //if
      end
      else begin
        if PacketState = psDeadAir then begin
          //Still dead.
        end
        else begin
          if PacketState = psPacket then begin
            //The packet is complete, send it on.
            if UnbufferedBits <> 0 then BufferByte;
            pDIOBuf.Timestamp := Now;
            PacketFIFO.Insert(pDIOBuf);
            pDIOBuf := nil;
          end;
          PacketState := psDeadAir;
        end;
      end;
    end;
    iPacketState := Ord(PacketState);
  end;
end;

{ TWritePortServer }

procedure TWritePortServer.ExecuteDIOWorker;
var
  DIOBuf: array of Word;
  pDIOBuf: PPacketData;
  I, L, PB, BytesTransferred, SubOff: LongInt;
  ThisByte: Byte;

begin
  pDIOBuf := nil;
  while not Terminated do begin
    if pDIOBuf <> nil then Dispose(pDIOBuf);
    pDIOBuf := PacketFIFO.Extract(INFINITE);
    if pDIOBuf = nil then Break;

    L := pDIOBuf.LenUsed;
    if L = 0 then Continue;

    PB := PacketBits;
    if Length(DIOBuf) <> PB then SetLength(DIOBuf, PB);
    ThisByte := 0; //For the compiler.
    for I := 0 to PB-1 do begin
      if (I and 7) = 0 then begin
        if I < L * 8 then
          ThisByte := pDIOBuf.Buf[I div 8]
        else
          ThisByte := 0
        ;
      end
      else
        ThisByte := ThisByte shr 1
      ;

      if (ThisByte and 1) <> 0 then DIOBuf[I] := OutDataMask else DIOBuf[I] := 0;
    end;

    L := PB * 2;
    SubOff := 0;
    repeat
      BytesTransferred := 0;
      Status := DIO_StreamFrame(DeviceIndex, L div 2, @DIOBuf[SubOff], LongWord(BytesTransferred));
      if Status = ERROR_HANDLE_EOF then begin
        Status := ERROR_SUCCESS; //A sub-block partial. We're streaming, so we don't really care.
      end
      else if Status <> ERROR_SUCCESS then begin
        bWorking := False;
        Exit; //Fail, kick out. `@ Maybe should re-find board and reinit, depending on error code?
      end;
      if BytesTransferred < 0 then BytesTransferred := 0; //Just in case.
      BytesTransferred := BytesTransferred and not 1; //Integral words.
      if BytesTransferred >= L then Break;
      L := L - BytesTransferred;
      SubOff := SubOff + (BytesTransferred div 2);
    until Terminated;
  end;
  SetLength(DIOBuf, 0);
  if pDIOBuf <> nil then Dispose(pDIOBuf);
end;

end.
